/*

*/

import renderer.scene.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;
import renderer.gui.*;

import java.awt.Color;
import java.awt.event.*;

/**
<pre>{@code
                    y         |
                    |         |
                    |         |
                    | v[1]    |
                  1 +         |
                    |         |     /
                    |         |   /
                    |         | /
                    |      -1 +--------------- image plane
                    |       /
                    |     /
                    |   /
                    | /                v[0]
               v[2] +-----------------+------> x
                  / 0                 1
                /
              /
            /
        1 +
        / v[2]
      /
    z
}</pre>
   Render a single wireframe triangle. This is just about
   the simplest possible model. It is useful for debugging.
*/
@SuppressWarnings("serial")
public class InteractiveTriangle_R5 extends InteractiveFrame
{
   private Scene scene;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
   */
   public InteractiveTriangle_R5(String title, int fbWidth, int fbHeight)
   {
      super(title, fbWidth, fbHeight);

      // Create a Model object to hold the geometry.
      Model model = new Model();

      // Create the geometry for the Model.
      Vertex v0 = new Vertex(1.0, 0.0, 0.0);
      Vertex v1 = new Vertex(0.0, 1.0, 0.0);
      Vertex v2 = new Vertex(0.0, 0.0, 0.0);
    //Vertex v2 = new Vertex(0.0, 0.0, 1.0);  // lands right on the camera

      // Give each vertex a different color.
      v0.setColor(new Color(255,  0,   0 )); // red
      v1.setColor(new Color( 0,  255,  0 )); // green
      v2.setColor(new Color( 0,   0,  255)); // blue

      // Add the geometry to the Model.
      model.addLineSegment(v0, v1);
      model.addLineSegment(v1, v2);
      model.addLineSegment(v2, v0);
/*
      // Or we can add the geometry to the model in this more efficient way.
      model.addVertex(v0, v1, v2);
      model.addLineSegment(0, 1);
      model.addLineSegment(1, 2);
      model.addLineSegment(2, 0);
*/
      // Push the model away from where the camera is.
      for (Vertex v : model.vertexList)
      {
         v.z -= 1;
      }


      // Create the Scene object that we shall render.
      scene = new Scene();

      // Add the Model to the Scene.
      scene.addModel(model);
   }


   // Implement part of the KeyListener interface.
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c)
      {
         Pipeline.debug = ! Pipeline.debug;
         Clip.debug = ! Clip.debug;
         Rasterize.debug = ! Rasterize.debug;
         RasterizeAntialias.debug = ! RasterizeAntialias.debug;
      }
      else if ('c' == c)
      {
         // Change the solid random color of the model.
         scene.modelList.get(0).setColorRandom();
      }
      else if ('e' == c)
      {
         // Change the random color of each vertex.
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.setColorRandom();
         }
      }
      else if ('E' == c)
      {
         // Change the random color of each line segment.
         for (LineSegment ls : scene.modelList.get(0).lineSegmentList)
         {
            ls.setColorRandom();
         }
      }
      else if ('a' == c)
      {
         Pipeline.doAntialiasing = ! Pipeline.doAntialiasing;
      }
      else if ('p' == c)
      {
         scene.camera.perspective = ! scene.camera.perspective;
      }
      else if ('s' == c) // Scale the triangle 10% smaller.
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.x /= 1.1;  // This does NOT work with
            v.y /= 1.1;  // perspective projection.
            v.z /= 1.1;
         }
      }
      else if ('S' == c) // Scale the triangle 10% larger.
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.x *= 1.1;  // This does NOT work with
            v.y *= 1.1;  // perspective projection.
            v.z *= 1.1;
         }
      }
      else if ('x' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.x -= 0.1;
         }
      }
      else if ('X' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.x += 0.1;
         }
      }
      else if ('y' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.y -= 0.1;
         }
      }
      else if ('Y' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.y += 0.1;
         }
      }
      else if ('z' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.z -= 0.1;
         }
      }
      else if ('Z' == c)
      {
         for (Vertex v : scene.modelList.get(0).vertexList)
         {
            v.z += 0.1;
         }
      }

      // Render again.
      FrameBuffer fb = this.fbp.getFrameBuffer();
      fb.clearFB(Color.black);
      Pipeline.render(scene, fb);
      fbp.update();
      repaint();
   }


   // Implement part of the ComponentListener interface.
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      // Get the new size of the FrameBufferPanel.
      int w = this.fbp.getWidth();
      int h = this.fbp.getHeight();

      // Create a new FrameBuffer that fits the new window size.
      FrameBuffer fb = new FrameBuffer(w, h);
      this.fbp.setFrameBuffer(fb);
      fb.clearFB(Color.black);
      Pipeline.render(scene, fb);
      fbp.update();
      repaint();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // Define initial dimensions for a FrameBuffer.
      int width  = 512;
      int height = 512;
      // Create an InteractiveFrame containing a FrameBuffer
      // with the given dimensions. NOTE: We need to call the
      // InteractiveTriangle_R5 constructor in the Java GUI Event
      // Dispatch Thread, otherwise we get a race condition
      // between the constructor (running in the main() thread)
      // and the very first ComponentEvent (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         new Runnable() // an anonymous inner class constructor
         {
            public void run() // implement the Runnable interface
            {
               // call the constructor that builds the gui
               new InteractiveTriangle_R5("Renderer 5", width, height);
            }
         }
      );
   }//main()


   private static void print_help_message()
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the triangle along the x, y, z axes.");
      System.out.println("Use the s/S keys to scale the size of the triangle.");
      System.out.println("Use the 'c' key to change the solid model color.");
      System.out.println("Use the 'e' key to change the random vertex colors.");
      System.out.println("Use the 'E' key to change the random line segment colors.");
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }
}
