/*

*/

package renderer.models;
import  renderer.scene.*;

/**
   Create a wireframe model of a regular dodecahedron
   with its center at the origin, having edge length
   <pre>{@code
     2*(sqrt(5)-1)/(1+sqrt(5)) = (1/2)*(sqrt(5)-1)^2 = 0.7639,
   }</pre>
   and with its vertices on a sphere of radius
   <pre>{@code
     2*sqrt(3)/(1+sqrt(5)) = 1.0705.
   }</pre>
<p>
   See <a href="https://en.wikipedia.org/wiki/Regular_dodecahedron" target="_top">
                https://en.wikipedia.org/wiki/Regular_dodecahedron</a>

   @see Tetrahedron
   @see Cube
   @see Octahedron
   @see Icosahedron
*/
public class Dodecahedron extends Model
{
   /**
      Create a regular dodecahedron with its center at
      the origin, having edge length
      <pre>{@code
        2*(sqrt(5)-1)/(1+sqrt(5)) = (1/2)*(sqrt(5)-1)^2 = 0.7639,
      }</pre>
      and with its vertices on a sphere of radius
      <pre>{@code
        2*sqrt(3)/(1+sqrt(5)) = 1.0705.
     }</pre>
   */
   public Dodecahedron()
   {
      super();

      // Create the dodecahedron's geometry.
      // It has 20 vertices and 30 edges.
      double t = (1 + Math.sqrt(5))/2;   // golden ratio
      double r = 1/t;
      double r2 = r * r;
      //https://en.wikipedia.org/wiki/Regular_dodecahedron#Cartesian_coordinates
      // (r, r, r)
      Vertex v00 = new Vertex(-r, -r, -r);
      Vertex v01 = new Vertex(-r, -r,  r);
      Vertex v02 = new Vertex(-r,  r, -r);
      Vertex v03 = new Vertex(-r,  r,  r);
      Vertex v04 = new Vertex( r, -r, -r);
      Vertex v05 = new Vertex( r, -r,  r);
      Vertex v06 = new Vertex( r,  r, -r);
      Vertex v07 = new Vertex( r,  r,  r);

      // (0, r2, 1)
      Vertex v08 = new Vertex( 0, -r2, -1);
      Vertex v09 = new Vertex( 0, -r2,  1);
      Vertex v10 = new Vertex( 0,  r2, -1);
      Vertex v11 = new Vertex( 0,  r2,  1);

      // (r2, 1, 0)
      Vertex v12 = new Vertex(-r2, -1,  0);
      Vertex v13 = new Vertex(-r2,  1,  0);
      Vertex v14 = new Vertex( r2, -1,  0);
      Vertex v15 = new Vertex( r2,  1,  0);

      // (1, 0, r2)
      Vertex v16 = new Vertex(-1,  0, -r2);
      Vertex v17 = new Vertex( 1,  0, -r2);
      Vertex v18 = new Vertex(-1,  0,  r2);
      Vertex v19 = new Vertex( 1,  0,  r2);
/*
      // These vertices create a dodecahedron with vertices
      // on a sphere of radius sqrt(3), and with edge length
      //    2/t = 4/(1 + sqrt(5)) = sqrt(5) - 1 = 1.2361.
      //https://en.wikipedia.org/wiki/Regular_dodecahedron#Cartesian_coordinates
      // (1, 1, 1)
      Vertex v00 = new Vertex(-1, -1, -1);
      Vertex v01 = new Vertex(-1, -1,  1);
      Vertex v02 = new Vertex(-1,  1, -1);
      Vertex v03 = new Vertex(-1,  1,  1);
      Vertex v04 = new Vertex( 1, -1, -1);
      Vertex v05 = new Vertex( 1, -1,  1);
      Vertex v06 = new Vertex( 1,  1, -1);
      Vertex v07 = new Vertex( 1,  1,  1);

      // (0, r, t)
      Vertex v08 = new Vertex( 0, -r, -t);
      Vertex v09 = new Vertex( 0, -r,  t);
      Vertex v10 = new Vertex( 0,  r, -t);
      Vertex v11 = new Vertex( 0,  r,  t);

      // (r, t, 0)
      Vertex v12 = new Vertex(-r, -t,  0);
      Vertex v13 = new Vertex(-r,  t,  0);
      Vertex v14 = new Vertex( r, -t,  0);
      Vertex v15 = new Vertex( r,  t,  0);

      // (t, 0, r)
      Vertex v16 = new Vertex(-t,  0, -r);
      Vertex v17 = new Vertex( t,  0, -r);
      Vertex v18 = new Vertex(-t,  0,  r);
      Vertex v19 = new Vertex( t,  0,  r);
*/
      // Create 30 line segments (that make up 12 faces).
//https://github.com/mrdoob/three.js/blob/master/src/geometries/DodecahedronGeometry.js
      addLineSegment(v03, v11);
      addLineSegment(v11, v07);
      addLineSegment(v07, v15);
      addLineSegment(v15, v13);
      addLineSegment(v13, v03);

      addLineSegment(v07, v19);
      addLineSegment(v19, v17);
      addLineSegment(v17, v06);
      addLineSegment(v06, v15);

      addLineSegment(v17, v04);
      addLineSegment(v04, v08);
      addLineSegment(v08, v10);
      addLineSegment(v10, v06);

      addLineSegment(v08, v00);
      addLineSegment(v00, v16);
      addLineSegment(v16, v02);
      addLineSegment(v02, v10);

      addLineSegment(v00, v12);
      addLineSegment(v12, v01);
      addLineSegment(v01, v18);
      addLineSegment(v18, v16);

      addLineSegment(v02, v13);

      addLineSegment(v18, v03);

      addLineSegment(v01, v09);
      addLineSegment(v09, v11);

      addLineSegment(v04, v14);
      addLineSegment(v14, v12);

      addLineSegment(v09, v05);
      addLineSegment(v05, v19);

      addLineSegment(v05, v14);
   }
}//Dodecahedron
