/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.util.List;
import java.util.ArrayList;

/**
    This renderer takes as its input a {@link Scene} data structure
    and a {@link FrameBuffer} data structure. This renderer mutates
    the {@link FrameBuffer}'s current viewport so that it is filled
    in with the rendered image of the scene represented by the
    {@link Scene} object.
<p>
    This implements our fifth rendering pipeline. This pipeline can
    process the colors stored in each {@link Vertex}. This renderer
    has the same four pipeline stages but {@link Vertex} color is
    interpolated by the {@link Clip} stage and {@link Vertex} color
    is also interpolated to pixels by the {@link Rasterize} stage.
*/
public class Pipeline
{
   public static boolean doAntialiasing = false;
   public static boolean debug = false;

   /**
      Mutate the {@link FrameBuffer}'s current viewport so that
      it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(Scene scene, FrameBuffer fb)
   {
      if ( debug )
         render_debug(scene, fb);
      else
         render_(scene, fb);
   }


   /**
      Renderer the {@link Scene} without any debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_(Scene scene, FrameBuffer fb)
   {
      // Render every Model in the Scene.
      for (Model model : scene.modelList)
      {
         if ( ! model.hidden )
         {
            // 0. Make a deep copy of the model.
            Model model2 = new Model(model);

            // 1. Apply the projection transformation.
            Projection.project(model2.vertexList, scene.camera);

            // 2. Clip each line segment to the standard view rectangle.
            List<LineSegment> lineSegmentList2 = new ArrayList<>();
            for (LineSegment ls : model2.lineSegmentList)
            {
               if ( Clip.clip(ls) )
               {
                  // Keep the line segments that are visible.
                  lineSegmentList2.add(ls);
               }
            }
            model2.lineSegmentList = lineSegmentList2;

            // 3. Apply viewport transformation to each visible line segment.
            boolean[] transformed = new boolean[model2.vertexList.size()];
            for (LineSegment ls : model2.lineSegmentList)
            {
               Viewport.viewport(ls, transformed, fb);
            }

            // 4. Rasterize each visible line segment into pixels.
            for (LineSegment ls : model2.lineSegmentList)
            {
               if ( doAntialiasing )
               {
                  RasterizeAntialias.rasterize(ls, fb);
               }
               else
               {
                  Rasterize.rasterize(ls, fb);
               }
            }
         }
      }
   }



   /**
      Render the {@link Scene} with quite a bit of debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_debug(Scene scene, FrameBuffer fb)
   {
      System.out.println("=== Render Scene ===================================");

      // Render every Model in the Scene.
      for (Model model : scene.modelList)
      {
         System.out.println("=== Render Model ===================================");

         if ( ! model.hidden )
         {
            // 0. Make a deep copy of the model.
            Model model2 = new Model(model);

            // log interesting information to standard output
            logVertexList("0. Input    ", model2.vertexList);

            // 1. Apply the projection transformation.
            Projection.project(model2.vertexList, scene.camera);

            // log interesting information to standard output
            logVertexList("1. Projected", model2.vertexList);

            // 2. Clip each line segment to the standard view rectangle.
            List<LineSegment> lineSegmentList2 = new ArrayList<>();
            for (LineSegment ls : model2.lineSegmentList)
            {
               // log interesting information to standard output
               logLineSegment("2. Clipping", ls);

               if ( Clip.clip(ls) )
               {
                  // log interesting information to standard output
                  logLineSegment("2. Clipped (accept)", ls);

                  // Keep the line segments that are visible.
                  lineSegmentList2.add(ls);
               }
               else
               {
                  // log interesting information to standard output
                  logLineSegment("2. Clipped (reject)", ls);
               }
            }
            model2.lineSegmentList = lineSegmentList2;

            // log interesting information to standard output
            logVertexList("2. Clipped  ", model2.vertexList);

            // 3. Apply viewport transformation to each visible line segment.
            boolean[] transformed = new boolean[model2.vertexList.size()];
            for (LineSegment ls : model2.lineSegmentList)
            {
               Viewport.viewport(ls, transformed, fb);

               // log interesting information to standard output
               logLineSegment("3. Viewport ", ls);
            }

            // 4. Rasterize each visible line segment into pixels.
            for (LineSegment ls : model2.lineSegmentList)
            {
               // log interesting information to standard output
               logLineSegment("4. Rasterize", ls);

               if ( doAntialiasing )
               {
                  RasterizeAntialias.rasterize(ls, fb);
               }
               else
               {
                  Rasterize.rasterize(ls, fb);
               }
            }
            System.out.println();
         }
         else
         {
            System.out.println("   Hidden model.");
         }
         System.out.println("=== End Model ===================================");
      }
      System.out.println("=== End Scene ===================================");
   }


   /**
      This method prints a {@link String} representation of the given vertex list.

      @param stage       name for the pipeline stage
      @param vertexList  vertex list whose string representation is to be printed
   */
   private static void logVertexList(String stage, List<Vertex> vertexList)
   {
      int i = 0;
      for (Vertex v : vertexList)
      {
         System.out.printf( stage + ": index = %3d, " + v.toString(), i );
         i++;
      }
   }


   /**
      This method prints a {@link String} representation of the given {@link LineSegment}.

      @param stage  name for the pipeline stage
      @param ls     {@link LineSegment} whose string representation is to be printed
   */
   private static void logLineSegment(String stage, LineSegment ls)
   {
      System.out.print( stage + ": " + ls.toString() );
      int index0 = ls.index[0];
      int index1 = ls.index[1];
      Vertex v0 = ls.model.vertexList.get(index0);
      Vertex v1 = ls.model.vertexList.get(index1);
      System.out.printf("    index = %3d, " + v0.toString(), index0);
      System.out.printf("                 " + v0.toStringColor());
      System.out.printf("    index = %3d, " + v1.toString(), index1);
      System.out.printf("                 " + v1.toStringColor(), index1);
   }
}
